// ////////////////////////////////////////////////////
//
// StdView.cpp : implementation of the CStdView class
// Based on class COGLView : public CView
// Inspect COGLView for standard OpenGL initialization
// Experimenting with Frenet-frames according to 
// Hill
//
////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Std.h"
#include "math.h"

#include "StdDoc.h"
#include "StdView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////
// You can put definitions, variables and helperfunctions
// here. Functions that does not rely on CStdView member
// variables or functions
#define TIMERID 3

//////////////////////////////////////////////////////////
// CStdView

IMPLEMENT_DYNCREATE(CStdView, COGLView)

BEGIN_MESSAGE_MAP(CStdView, COGLView)
	//{{AFX_MSG_MAP(CStdView)
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_COMMAND(ID_SHOW_CYLINDERS, OnShowCylinders)
	ON_UPDATE_COMMAND_UI(ID_SHOW_CYLINDERS, OnUpdateShowCylinders)
	ON_COMMAND(ID_SHOW_ANIMATION, OnShowAnimation)
	ON_UPDATE_COMMAND_UI(ID_SHOW_ANIMATION, OnUpdateShowAnimation)
	ON_WM_TIMER()
	ON_COMMAND(ID_SHOW_AXES, OnShowAxes)
	ON_UPDATE_COMMAND_UI(ID_SHOW_AXES, OnUpdateShowAxes)
	ON_COMMAND(ID_SHOW_RUN, OnShowRun)
	ON_UPDATE_COMMAND_UI(ID_SHOW_RUN, OnUpdateShowRun)
	ON_COMMAND(ID_SHOW_BEZIER, OnShowBezier)
	ON_UPDATE_COMMAND_UI(ID_SHOW_BEZIER, OnUpdateShowBezier)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

///////////////////////////////////////////////////////////
// CStdView construction/destruction

CStdView::CStdView()
{
		
	// set up beziercurve's conttrolpoints
	bgp0.x=-10.0f;	bgp0.y=-10.0f;	bgp0.z=0.0f;
	bgp1.x=-10.0f;	bgp1.y=20.0f;	bgp1.z=0.0f;
	bgp2.x=10.0f;	bgp2.y=-30.0f;	bgp2.z=0.0f;
	bgp3.x=20.0f;	bgp3.y=-10.0f;	bgp3.z=0.0f;

	
	// fill the btab with weights T x M (independant of controlpoints)
	// for cubic curve. This is the weigth-functions
	// Bezier: btab (position)
	int tix=0;
	while(tix<PCNT)
	{
		GLfloat t=(1.0f/(1.0f*PCNT))*(1.0f*tix);
		btab[0][tix]=(1-t)*(1-t)*(1-t);
		btab[1][tix]=3.0f*t*(1-t)*(1-t);
		btab[2][tix]=3.0f*t*t*(1-t);
		btab[3][tix]=t*t*t;
		tix++;
	}
	// The derivative of Bezier: d_btab (direction)
	tix=0;
	while(tix<PCNT)
	{
		GLfloat t=(1.0f/(1.0f*PCNT))*(1.0f*tix);
		d_btab[0][tix]=-3.0f*t*t+6.0f*t-3.0f;
		d_btab[1][tix]=9.0f*t*t-12.0f*t+3.0f;
		d_btab[2][tix]=-9.0f*t*t+6.0f*t;
		d_btab[3][tix]=3.0f*t*t;
		tix++;
	}

	// The double derivative of Bezier: (change of direction)
	tix=0;
	while(tix<PCNT)
	{
		GLfloat t=(1.0f/(1.0f*PCNT))*(1.0f*tix);
		dd_btab[0][tix]=-6.0f*t+6.0f;
		dd_btab[1][tix]=18.0f*t-12.0f;
		dd_btab[2][tix]=-18.0f*t+6.0f;
		dd_btab[3][tix]=6.0f*t;
		tix++;
	}
	
	// done here and each time controlpoints are
	// changed, on timer-event
	makeBezier(bgp0,bgp1,bgp2,bgp3);
	make_d_Bezier(bgp0,bgp1,bgp2,bgp3);
	make_dd_Bezier(bgp0,bgp1,bgp2,bgp3);


	// and the mousing
	vx=0.0f;
	vy=0.0f;
	is_moving=false;

	// which position to make a golden ring in
	showix=1;
	stepix=1;

	// what we will see
	bShowCylinders=FALSE;
	bAnimate=FALSE;
	bShowAxes=FALSE;
	bRun=FALSE;
	bDrawBezier=true;

	m_hRC=NULL;
	m_hDC=NULL;

};

CStdView::~CStdView()
{}

////////////////////////////////////////////////////////////
// CStdView diagnostics
// You should usually not touch these:
#ifdef _DEBUG
void CStdView::AssertValid() const
{COGLView::AssertValid();}
void CStdView::Dump(CDumpContext& dc) const
{COGLView::Dump(dc);}

CStdDoc* CStdView::GetDocument() // non-debug version is inline
{	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CStdDoc)));
	return (CStdDoc*)m_pDocument;
}
#endif //_DEBUG

//////////////////////////////////////////////////////////////
// CStdView operations

void CStdView::DrawScene()
{
	
	// Clear the color and depth buffers.
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
	// Set up for scene
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// position the eye relative to scene
	gluLookAt(	0.0f,0.0f,40.0f,	// eye
				0.0f,0.0f,0.0f,	// looking at
				0.0f,1.0f,0.0f);// is up

	//mousing
	glRotatef(vx,1.0f,0.0f,0.0f);
	glRotatef(vy,0.0f,1.0f,0.0f);

	// Bezier, Derivative of Bezier and 
	// double derivative are all prepared
	// See: OnTimer and calculation of 
	// weigth functions in the constructor

	// the Bezier function
	if(bDrawBezier)
	{
		// draw the bezierfunction strait
		CMaterial::SetMaterial(RED_PLASTIC);
		glLineWidth(4);
		drawBezier();
	}

	// hold on to your hat
	// do the t-walk and draw in frenet frames
	int ix=0;
	while(ix<PCNT)
	{
		
		// Bezier is now tabulated in:
		//	(bez[ix][0],bez[ix][1],bez[ix][2])
		// The derivative (tangent) is tabulated in:
		//	(d_bez[ix][0],d_bez[ix][1],d_bez[ix][2])
		// The double derivated is tabulated in:
		//	(dd_bez[ix][0],dd_bez[ix][1],dd_bez[ix][2])

		// finding the normalized tangent: 
		// T pointing along the curve
		v3d T(d_bez[ix][0],d_bez[ix][1],d_bez[ix][2]);
		T.normalize();

		// finding the normalized d cross dd vector: 
		// B pointing perpendicular to T
		v3d B(d_bez[ix][0],d_bez[ix][1],d_bez[ix][2]);
		B.cross(dd_bez[ix][0],dd_bez[ix][1],dd_bez[ix][2]);
		//B.normalize();
		
		// if we are not happy with the doublederivative 
		// as basis for the second vector
		// maybe we can use an other vector here ?
		// what is the effect of a fixed direction ?

		//B.cross(0.0f,0.0f,1.0f);
		B.normalize();

		// finding a third perpendiculat vector: 
		// N perpendicular to both T and B
		v3d N(B.x,B.y,B.z);
		N.cross(T);  // and it is normalized since B and T are

		// we want to move to the correct point on the Beziercurve
		// as tabulated in (bez[ix][0],bez[ix][1],bez[ix][2])
		v3d C(bez[ix][0],bez[ix][1],bez[ix][2]);
		
		// at this point T,B,N describes the coordinate system
		// we will use at the position C 
		// N will act as x-axes
		// B will act as y-axes
		// T will act as z-axes

		// setting up the matix that will take us to the
		// Frenet frame located at this t-point
		// M=|N,B,T,C|, where C is the Bezier itself
		// note transpose of matrix

		GLfloat M[16]={
			B.x,B.y,B.z,0,
			N.x,N.y,N.z,0,
			T.x,T.y,T.z,0,
			C.x,C.y,C.z,1
		};
		
		// take care of the matrix we have
		glPushMatrix();
		
		//multiply our newly made Frenet matrix to OpenGLs current matrix
		glMultMatrixf(&M[0]);

		// and we are ready to draw what ever we want to draw

		// axes
		if(bShowAxes)
		{
			// draw the axis
			glLineWidth(2);
			CMaterial::SetMaterial(RED_PLASTIC);
			glBegin(GL_LINES);
				glVertex3f(0.0f,0.0f,0.0f);
				glVertex3f(0.0f,0.0f,2.0f);
			glEnd();
			CMaterial::SetMaterial(GREEN_PLASTIC);
			glBegin(GL_LINES);
				glVertex3f(0.0f,0.0f,0.0f);
				glVertex3f(0.0f,2.0f,0.0f);
			glEnd();
			CMaterial::SetMaterial(BLACK_PLASTIC);
			glBegin(GL_LINES);
				glVertex3f(0.0f,0.0f,0.0f);
				glVertex3f(2.0f,0.0f,0.0f);
			glEnd();
		}
		
		
		// the green tubes
		if(bShowCylinders)
		{
			// draw a sylinder as a temporary solution
			GLUquadricObj *glpQ=gluNewQuadric();
			CMaterial::SetMaterial(EMERALD);
			gluCylinder(glpQ,1.1f,1.1f,1.0f,15,1);
			gluDeleteQuadric(glpQ);
		}
		
		// the golden tube
		if((bRun)&&(ix==showix))
		{
			GLUquadricObj *glpQ=gluNewQuadric();
			CMaterial::SetMaterial(GOLD);
			gluCylinder(glpQ,1.3f,1.3f,1.0f,15,1);
			gluDeleteQuadric(glpQ);
		}


		glPopMatrix();
		
		// next t-pos
		ix++;
	}
	

	// finnish drawing
	glFlush();
}

void CStdView::InitializeRContext()
{
	// Replace this call with your own one-time 
	// initialization code
	COGLView::InitializeRContext();
}

void CStdView::SetPerspective(int cx, int cy)
{
    // We will work on the projection matrix,
	// so set matrix mode
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	glOrtho(-30.0f,30.0f,-30.0f*cy/cx,30.0f*cy/cx,-100.f,100.0f);
	// set up your viewing volume
	// gluPerspective(70.0f, (GLdouble)cx/cy, 0.0f, 70.0f);
	// reset matrix mode 
	glMatrixMode(GL_MODELVIEW);
}

void CStdView::drawBezier()
{
	// hand draw of bezier function
	// everything calculated in beforehand
	glBegin(GL_LINE_STRIP);
	int ix=0;
	while(ix<PCNT)
	{
		glVertex3f(bez[ix][0],bez[ix][1],bez[ix][2]);
		ix++;
	}
	glEnd();
}

//**************************************************************
// calculate tables for function, derivative and doublederivative
// The respective weigth-functions are already prepared in
// btab, d_btab and dd_btab

void CStdView::makeBezier(v3d p0,v3d p1,v3d p2,v3d p3)
{
	int ix=0;
	while(ix<PCNT)
	{
		bez[ix][0]=p0.x*btab[0][ix]+p1.x*btab[1][ix]+p2.x*btab[2][ix]+p3.x*btab[3][ix];
		bez[ix][1]=p0.y*btab[0][ix]+p1.y*btab[1][ix]+p2.y*btab[2][ix]+p3.y*btab[3][ix];
		bez[ix][2]=p0.z*btab[0][ix]+p1.z*btab[1][ix]+p2.z*btab[2][ix]+p3.z*btab[3][ix];
		
		ix++;
	}
}
void CStdView::make_d_Bezier(v3d p0,v3d p1,v3d p2,v3d p3)
{
	int ix=0;
	while(ix<PCNT)
	{
		d_bez[ix][0]=p0.x*d_btab[0][ix]+p1.x*d_btab[1][ix]+p2.x*d_btab[2][ix]+p3.x*d_btab[3][ix];
		d_bez[ix][1]=p0.y*d_btab[0][ix]+p1.y*d_btab[1][ix]+p2.y*d_btab[2][ix]+p3.y*d_btab[3][ix];
		d_bez[ix][2]=p0.z*d_btab[0][ix]+p1.z*d_btab[1][ix]+p2.z*d_btab[2][ix]+p3.z*d_btab[3][ix];
		
		ix++;
	}
}

void CStdView::make_dd_Bezier(v3d p0,v3d p1,v3d p2,v3d p3)
{
	int ix=0;
	while(ix<PCNT)
	{
		dd_bez[ix][0]=p0.x*dd_btab[0][ix]+p1.x*dd_btab[1][ix]+p2.x*dd_btab[2][ix]+p3.x*dd_btab[3][ix];
		dd_bez[ix][1]=p0.y*dd_btab[0][ix]+p1.y*dd_btab[1][ix]+p2.y*dd_btab[2][ix]+p3.y*dd_btab[3][ix];
		dd_bez[ix][2]=p0.z*dd_btab[0][ix]+p1.z*dd_btab[1][ix]+p2.z*dd_btab[2][ix]+p3.z*dd_btab[3][ix];

		ix++;
	}
}

//***********************************************************
// mousing

void CStdView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	last_p=point;
	is_moving=true;
	SetCapture();
	
	CView::OnLButtonDown(nFlags, point);
}

void CStdView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	ReleaseCapture();
	is_moving=false;
	CView::OnLButtonUp(nFlags, point);
}

void CStdView::OnMouseMove(UINT nFlags, CPoint point) 
{
	if((is_moving)&&((abs(last_p.x-point.x)>2)||(abs(last_p.y-point.y)>2)))
	{
		// then we find an intuitive rot angle
		CRect R;
		GetClientRect(&R);
		
		vx+=90.0f*((GLfloat (point.y-last_p.y))/GLfloat (R.Height()));
		vy+=90.0f*((GLfloat (point.x-last_p.x))/GLfloat (R.Width()));
		
		// rememeber point and force immediate redraw 
		last_p=point;
		RedrawWindow();

	}
	
	CView::OnMouseMove(nFlags, point);
}

void CStdView::OnShowCylinders() 
{
	bShowCylinders=!bShowCylinders;
	RedrawWindow();
	
}

void CStdView::OnUpdateShowCylinders(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(bShowCylinders);
	
}

void CStdView::OnShowAnimation() 
{
	if(!bAnimate)
	{
		bAnimate=TRUE;
		SetTimer(TIMERID,50,NULL);
	}
	else
	{
		bAnimate=FALSE;
		KillTimer(TIMERID);
	}
	
}

void CStdView::OnUpdateShowAnimation(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(bAnimate);	
}

void CStdView::OnTimer(UINT nIDEvent) 
{
	static GLfloat dy1=-1.0f;
	static GLfloat dy2=0.5f;
	static GLfloat dz0=-0.5f;
	static GLfloat dz3=1.0f;
	if(bAnimate)
	{
		// adjust controlpoints of 
		// the Bezierfunction and force update
		// actual values to change are:

		bgp1.y+=dy1;
		bgp2.y+=dy2;
		if((bgp1.y>30.0f)||(bgp1.y<-30.0f)) dy1*=-1.0f;
		if((bgp2.y>30.0f)||(bgp2.y<-30.0f)) dy2*=-1.0f;

		bgp0.z+=dz0;
		bgp3.z+=dz3;
		if((bgp0.z>20.0f)||(bgp0.z<-20.0f)) dz0*=-1.0f;
		if((bgp3.z>20.0f)||(bgp3.z<-20.0f)) dz3*=-1.0f;

		// mark a point moving back and forth on 
		// the bezier
		if(showix==PCNT)
			stepix=-1;
		else if(showix==0)
			stepix=1;
		showix+=stepix;

		makeBezier(bgp0,bgp1,bgp2,bgp3);
		make_d_Bezier(bgp0,bgp1,bgp2,bgp3);
		make_dd_Bezier(bgp0,bgp1,bgp2,bgp3);


		RedrawWindow();
	}

	
	COGLView::OnTimer(nIDEvent);
}

void CStdView::OnShowAxes() 
{
	bShowAxes=!bShowAxes;
	RedrawWindow();
	
}

void CStdView::OnUpdateShowAxes(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(bShowAxes);
}

void CStdView::OnShowRun() 
{
	bRun=!bRun;
	RedrawWindow();
}

void CStdView::OnUpdateShowRun(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(bRun);
	
}

void CStdView::OnShowBezier() 
{
	bDrawBezier=!bDrawBezier;
	RedrawWindow();
}

void CStdView::OnUpdateShowBezier(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(bDrawBezier);
	
}
